//------------------------------------------------------------------------------
// Силовой модуль. Контроллер ТЭНа
// Список изменений - в конце файла
// OldBean, 03.02.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>

#define F_CPU 16000000UL
#define ADDR 0x13      // Адрес устройства на шине I2C
#define PULSE_WIDTH 78 // Длительность управляющего импульса - около 5 мс
                         
volatile uint8_t pdm = 0;   // Требуемый уровень PDM-модуляции в %
volatile int16_t err = 0;   // Ошибка дискретизации
volatile int8_t pp = 1;     // Относительный знак предстоящего полупериода
                            // 1 - положительный, 0 - отрицательный
volatile int16_t ps = 0;    // Отклонение постоянной составляющей от 0
volatile uint8_t pFlag = 0; // Флаг: 1 - пропускаем полупериод к нагрузке, т.е.
                            // подаем управляющий импульс на моську, 0 - нет
volatile int16_t lev;       // Текущий уровень (pdm + err)
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  switch(TW_STATUS) {    // Статус - в 5 старших битах регистра TWSR
    case TW_SR_DATA_ACK: // Данные приняты, ACK отправлен - 0x80
      pdm = TWDR;
      if(pdm > 100) pdm = 100;
      break;
    case TW_ST_SLA_ACK:  // Запрос SLA+R, ACK отправлен - 0xA8
      break;
    case TW_BUS_ERROR:   // Неверные условия "старт" или "стоп" - 0x00
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN); // "Сброс" TWI
      break;
  }
  TWCR |= _BV(TWINT);   // Флаг TWINT здесь НЕ очищается автоматически! Очистим.
} 
//------------------------------------------------------------------------------
ISR(INT0_vect) { // Внешнее прерывание INT0 (Zero от детектора нуля)
  lev = pdm + err; // Текущий уровень выходного сигнала (мощности)
  pFlag = (lev >= 50)? 1 : 0; // Предварительное решение
  if(ps*pFlag*pp > 0)         // Однако, если подадим импульс, то постоянная
    pFlag = 0;                // составляющая возрастет. Значит - не будем
                              // подавать - отложим да следующего полупериода
  if(pFlag) { // Положительное решение принято - дадим этот полупериод нагрузке
    PORTD |= _BV(PD0);        // Подаем управляющий импульс
    err = lev - 100;          // Оцениваем ошибку (сколько "переложили")
  } else err = lev;           // Если не дали - сколько "недоложили"
  GTCCR = _BV(PSRASY);        // Сброс предделителя таймера 2
  TCNT2 = 0;                  // Обнуляем счетчик таймера 2
}
//------------------------------------------------------------------------------
ISR(TIMER2_COMPA_vect) { // Обработчик по совпадению с регистром A
  PORTD &= ~_BV(PD0); // Завершаем управляющий импульс
}
//------------------------------------------------------------------------------
int main(void) {
  DDRD = _BV(PD0); // Пин PD0 - на вывод (управление оптосимистором)
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация внешнего прерывания INT0
  EICRA = _BV(ISC01) | _BV(ISC00); // Прерывания по фронту на пине PD2/INT0
  EIMSK = _BV(INT0); // Разрешаем внешние прерывания на пине PD2/INT0
  // Инициализация таймера 2 для формирования управляющего импульса для
  // MOC3083 длительностью приблизительно 5 мс
  // Счетчик работает в обычном режиме (Normal mode)
  TCCR2B |= _BV(CS22) | _BV(CS21) | _BV(CS20); // Предделитель на 1024
  OCR2A = PULSE_WIDTH; // Длительность управляющего импульса
  TIMSK2 = _BV(OCIE2A); // Разрешаем прерывания по совпадению
  
  sei(); // Разрешаем глобальные прерывания
  while(1) { // Бесконечный цикл
  }
  return 0;
}
//------------------------------------------------------------------------------
// Список изменений
// 1. 03.02.2018 - Отлажен базовый вариант. Разрешение PDM-модуляции - 1%
//------------------------------------------------------------------------------

